#!/usr/bin/env python3
"""
Live YAML editor for fluidpatcher banks.

Edits in the text pane are parsed immediately. Discovered patches
are added to the Patch menu, and can be applied to audition changes
to the bank. Bank errors are shown in the status bar and dumped to
stdout while editing - this is diagnostic and non-breaking.

Note: this editor can be used alongside a DAW.
Connect MIDI/audio to fluidsynth.
"""

from pathlib import Path
import tkinter as tk
from tkinter import ttk, filedialog
import traceback

from yaml import YAMLError

import fluidpatcher
from fluidpatcher.bankfiles import BankSyntaxError, BankValidationError

BANKS_PATH = fluidpatcher.CONFIG["banks_path"]

class BankEditorApp(tk.Tk):
    def __init__(self):
        super().__init__()
        self.title("FluidPatcher Live Bank Editor")
        self.geometry("900x600")
        self.lastfile = ""
        self._parse_after_id = None

        # create menus
        menubar = tk.Menu(self)
        filemenu = tk.Menu(menubar, tearoff=False)
        filemenu.add_command(
            label="Open", underline=0, command=self.open_file
        )
        filemenu.add_command(
            label="Save", underline=0, command=self.save_file
        )
        filemenu.add_separator()
        filemenu.add_command(
            label="Quit", underline=0, command=self.quit
        )
        self.patchmenu = tk.Menu(menubar, tearoff=False)
        menubar.add_cascade(
            label="File", menu=filemenu, underline=0
        )
        menubar.add_cascade(
            label="Patches", menu=self.patchmenu, underline=0
        )
        self.config(menu=menubar)

        # create widgets
        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)
        self.text = tk.Text(
            self,
            wrap="none",
            undo=True,
            font="TkFixedFont",
        )
        self.text.grid(row=0, column=0, sticky="nsew")
        yscroll = ttk.Scrollbar(self, orient="vertical", command=self.text.yview)
        xscroll = ttk.Scrollbar(self, orient="horizontal", command=self.text.xview)
        yscroll.grid(row=0, column=1, sticky="ns")
        xscroll.grid(row=1, column=0, sticky="ew")
        self.text.configure(yscrollcommand=yscroll.set,
                            xscrollcommand=xscroll.set)
        bar = ttk.Frame(self)
        bar.grid(row=2, column=0, columnspan=2, sticky="ew")
        bar.columnconfigure(0, weight=1)
        self.status_left = ttk.Label(bar, text="Ready", anchor="w")
        self.status_right = ttk.Label(bar, text="Ln 1, Col 1", anchor="e")
        self.status_left.grid(row=0, column=0, sticky="ew", padx=4)
        self.status_right.grid(row=0, column=1, sticky="e", padx=4)

        # add binds
        self.text.bind("<<Modified>>", self._on_text_modified)
        self.text.bind("<KeyRelease>", self._update_cursor_pos)
        self.text.bind("<ButtonRelease>", self._update_cursor_pos)

    def open_file(self):
        f = filedialog.askopenfilename(
            initialdir=self.lastfile.parent if self.lastfile else BANKS_PATH,
            initialfile=self.lastfile.name if self.lastfile else "",
            filetypes=[("YAML files", "*.yaml *.yml"), ("All files", "*.*")],
        )
        if not f:
            return
        self.lastfile = Path(f)
        content = self.lastfile.read_text()
        self.text.delete("1.0", "end")
        self.text.insert("1.0", content)
        self._update_cursor_pos()

    def save_file(self):
        f = filedialog.asksaveasfilename(
            initialdir=self.lastfile.parent if self.lastfile else BANKS_PATH,
            initialfile=self.lastfile.name if self.lastfile else "",
            defaultextension=".yaml",
            filetypes=[("YAML files", "*.yaml *.yml"), ("All files", "*.*")]
        )
        if not f:
            return
        self.lastfile = Path(f)
        self.lastfile.write_text(self.text.get("1.0", "end-1c"))
        self.set_status(f"Saved {f}")

    def _on_text_modified(self, e=None):
        self.text.edit_modified(False)
        if self._parse_after_id is not None:
            self.after_cancel(self._parse_after_id)
        self._parse_after_id = self.after(300, self.parse_bank)

    def _update_cursor_pos(self, e=None):
        line, col = self.text.index("insert").split(".")
        self.status_right.config(text=f"Ln {line}, Col {int(col) + 1}")

    def parse_bank(self):
        self._parse_after_id = None
        text = self.text.get("1.0", "end-1c")
        try:
            fp.load_bank(raw=text)
        except Exception as e:
            if isinstance(e, BankSyntaxError) and e.mark:
                buflines = e.mark.buffer.splitlines()
                for i, line in enumerate(text.splitlines()):
                    if line in buflines:
                        if buflines.index(line) == e.mark.line:
                            break
                        if buflines.index(line) > e.mark.line:
                            i -= 1
                            break
                self.set_status(f"{type(e).__name__}: {e.msg} on line {i + 1}")
            else:
                self.set_status(f"{type(e).__name__}: {e}")
            traceback.print_exception(type(e), e, e.__traceback__)
            self.status_left.configure(foreground="red")
        else:
            self.status_left.configure(foreground="black")
            self.patchmenu.delete(0, self.patchmenu.index('end'))
            for i, name in enumerate(fp.bank.patches):
                self.patchmenu.add_command(
                    label=name,
                    command=lambda n=name: self.select_patch(n)
                )
            self.set_status("Bank parsed successfully")

    def select_patch(self, name):
        fp.apply_patch(name)
        self.set_status(f"Applied patch {name}")

    def set_status(self, msg: str):
        self.status_left.config(text=msg.replace("\n", " "))

fp = fluidpatcher.FluidPatcher()

def main():
    app = BankEditorApp()
    app.mainloop()

if __name__ == "__main__":
    main()





"""
Bank file parsing and validation utilities for FluidPatcher.

This module defines:
  - Custom YAML tags for expressing patches, rules, sequences, and other objects
  - Loader and dumper wrappers for FluidPatcher’s extended YAML dialect
  - Structured exceptions distinguishing syntax vs. semantic errors
  - A Bank class that represents a parsed, validated bank file
  - Helper functions for name resolution and tree walking

It enables FluidPatcher to treat YAML as a rich declarative language for
describing patches, MIDI routing, sequencing, and effects.
"""

from copy import deepcopy
import re

import yaml


TYPE_ALIAS = {}
for alts in [
    ["note", "note_on", "nt"],
    ["ctrl", "control_change", "cc"],
    ["prog", "program_change", "pc"],
    ["pbend", "pitchwheel", "pb"],
    ["cpress", "aftertouch", "cp"],
    ["kpress", "polytouch", "kp"],
    ["sysex"],
    ["clock"],
    ["start"],
    ["continue"],
    ["stop"],
]:
    for a in alts:
        TYPE_ALIAS[a] = alts[0]

scinote = re.compile(r"([+-]?)([A-G])([b#]?)(-?[0-9])")

def resolve(s, names={}):
    """
    Resolve a YAML scalar into a usable numeric value.

    Resolution rules:
      - Replace symbolic names using the ``names`` mapping
      - Convert scientific pitch notation (e.g. ``C#4`` → MIDI note 61)
      - Parse strings to integers or floats

    Args:
      s: String, integer, float, or name reference
      names (dict): Optional mapping of symbolic names to values

    Returns:
      int | float

    Raises:
      BankValidationError: if ``s`` cannot be interpreted or resolved
    """
    if s in names:
        s = names[s]
    if isinstance(s, str):
        if sci := scinote.match(s):
            sci = sci.groups("")
            sign = -1 if sci[0] == "-" else 1
            note = "C D EF G A B".find(sci[1])
            acc = ["b", "", "#"].index(sci[2]) - 1
            octave = int(sci[3])
            s = sign * ((octave + 1) * 12 + note + acc)
        else:
            try:
                s = float(s)
            except (ValueError, TypeError):
                raise BankValidationError(f"Unresolved name '{s}'")
            else:
                if s.is_integer():
                    s = int(s)
    return s

def _walk(node, path=(), names={}):
    if node is None:
        raise BankValidationError(
            "Null value in bank", path
        )
    if hasattr(node, "_validate"):
        try:
            node._validate(path, names)
        except BankValidationError as e:
            e.path = path + e.path[len(path):]
            raise(e)
        except Exception as e:
            raise BankValidationError(str(e), path)
    elif isinstance(node, dict):
        for k, v in node.items():
            _walk(v, path + (k,), names)
    elif isinstance(node, list):
        for i, v in enumerate(node):
            _walk(v, path + (i,), names)


class BankError(Exception):
    """Base class for all bank parsing and validation errors."""
    pass
    

class BankSyntaxError(BankError):
    """
    Error raised when the YAML text is malformed and cannot be parsed.

    Attributes:
      msg (str): Reason for failure
      mark (tuple): YAML mark object describing the error
    """
    def __init__(self, msg):
        self.msg = msg
        self.mark = None

    @classmethod
    def from_yamlexc(cls, exc):
        """
        Wraps a YAMLError and extracts info about
        the location and reason
        """
        obj = cls.__new__(cls)
        obj.msg = (
            getattr(exc, "context", None) or
            getattr(exc, "problem", None) or
            str(exc)
        )
        obj.mark = (
            getattr(exc, "context_mark", None) or
            getattr(exc, "problem_mark", None) or
            None
        )
        return obj

    def __str__(self):
        if self.mark:
            loc = f" at line {self.mark.line}, column {self.mark.column}"
        else:
            loc = ""
        return self.msg + loc    


class BankValidationError(BankError):
    """
    Raised when a bank file is valid YAML but contains incorrect or
    unsupported values.

    Attributes:
      msg (str): Reason for failure
      path (tuple): Hierarchical path to the failing node
    """
    def __init__(self, msg, path=()):
        self.msg = msg
        self.path = path

    def __str__(self):
        if self.path:
            loc = " in " + ".".join(map(str, self.path))
        else:
            loc = ""
        return self.msg + loc


class BankLoader(yaml.SafeLoader):
    """
    YAML loader extended with:
      - Custom tags for bank object types
      - Path and implicit resolvers
      - Construction helpers used by Bank()
    """
    pass


class BankDumper(yaml.SafeDumper):
    """
    YAML dumper that emits FluidPatcher’s extended object syntax and
    preserves key order and structure wherever possible.
    """
    pass


class Bank:
    """
    Parsed representation of a FluidPatcher bank YAML document.

    Attributes:
      root (dict):
        Raw root-level configuration data.
      patch (dict):
        Raw dictionary of patches (unmerged).
      patches (list[str]):
        Patch names in declaration order.
      soundfonts (list[SFPreset]):
        Dynamic list of soundfont files required by patches.
    """
    def __init__(self, text):
        try:
            self.root = yaml.load(text, Loader=BankLoader)
        except yaml.YAMLError as e:
            raise BankSyntaxError.from_yamlexc(e)
        self.patch = self.root.setdefault("patches", {})
        names = self.root.get("names", {})
        _walk(self.root, path=(), names=names)

    @property
    def patches(self):
        return list(self.patch)

    @property
    def soundfonts(self):
        sfonts = set()
        for zone in self:
            for item in zone.values():
                if isinstance(item, SFPreset):
                    sfonts.add(item.file)
        return sfonts

    def __getitem__(self, name):
        return _Patch(self.root, self.patch[name])

    def __setitem__(self, name, p):
        self.patch[name] = p

    def __delitem__(self, name):
        del self.patch[name]

    def __contains__(self, name):
        return name in self.patch

    def __len__(self):
        return len(self.patches)

    def __iter__(self):
        return iter([self.root, *self.patch.values()])

    def index(self, name):
        return list(self.patch).index(name)

    def dump(self):
        bank = self.root | {"patches": self.patch}
        return yaml.dump(bank, Dumper=BankDumper, sort_keys=False)


class _Patch:
    """
    Lightweight merged view over root and per-patch data.

    Lookup rules:
      - Scalar fields fall back to root when absent in patch
      - Lists (rules, messages) concatenate root + patch values
      - Dict-like structures merge root defaults with overrides

    Created automatically via Bank.__getitem__.
    """

    def __init__(self, root, patch):
        self._patch = patch
        self._root = root
        
    def __getitem__(self, name):    
        if isinstance(name, int):
            return self._patch.get(name) or self._root.get(name)
        elif name in ("rules", "messages"):
            return self._root.get(name, []) + self._patch.get(name, [])
        else:
            return self._root.get(name, {}) | self._patch.get(name, {})

    def copy(self, **pars):
        return deepcopy(self._patch | pars)


class SFPreset(yaml.YAMLObject):
    """
    SoundFont preset reference

    Attributes:
      file (str): Path to the .sf2 file
      bank (int): SoundFont bank index
      prog (int): Program index
    """
    yaml_tag = "!sfpreset"
    yaml_loader = BankLoader
    yaml_dumper = BankDumper
    yaml_regex = re.compile(r"^(.+\.sf2):(\d+):(\d+)$", flags=re.I)
    zone = None

    def __init__(self, file, bank, prog):
        self.file = file
        self.bank = int(bank)
        self.prog = int(prog)

    @classmethod
    def from_yaml(cls, loader, node):
        return cls(*loader.construct_scalar(node).split(":"))

    @classmethod
    def to_yaml(cls, dumper, data):
        return dumper.represent_scalar(cls.yaml_tag, str(data))

    def __repr__(self):
        return f"{self.file}:{self.bank:03d}:{self.prog:03d}"


class MidiMessage(yaml.YAMLObject):
    """
    Description of a single MIDI message

    Attributes:
      type (str): Message type (“note”, “ctrl”, “sysex”, ...)
      chan (int): Channel number (where applicable)
      num (int): Note/controller number (where applicable)
      val (int): Value (velocity, CC value, pitch bend, etc.)
    """
    yaml_tag = "!midimsg"
    yaml_loader = BankLoader
    yaml_dumper = BankDumper
    yaml_regex = re.compile(rf"^({'|'.join(TYPE_ALIAS)}):\S*$")
    zone = "messages"
    zone_type = list

    def __init__(self, **pars):
        self.__dict__.update(pars)
        if "_text" not in pars:
            match pars:
                case {"type": "sysex", "val": data}:
                    self._text = ":".join(["sysex"] + [str(b) for b in data])
                case {"type": type, "chan": chan, "num": num, "val": val}:
                    self._text = f"{type}:{chan}:{num}:{val}"
                case {"type": type, "chan": chan, "val": val}:
                    self._text = f"{type}:{chan}:{val}"
                case {"type": type}:
                    self._text = type
        self._validate()

    @classmethod
    def from_yaml(cls, loader, node):
        text = loader.construct_scalar(node)
        match [x for x in text.split(":") if x]:
            case ["sysex", *data]:
                pars = dict(type="sysex", val=data)
            case [type, chan, num, val]:
                pars = dict(type=type, chan=chan, num=num, val=val)
            case [type, chan, val]:
                pars = dict(type=type, chan=chan, val=val)
            case [type]:
                pars = dict(type=type, _text=text)
        obj = cls.__new__(cls)
        obj.__dict__.update(pars)
        obj._pars = pars
        obj._text = text
        return obj
        
    @classmethod
    def to_yaml(cls, dumper, data):
        return dumper.represent_scalar(cls.yaml_tag, str(data))
            
    def copy(self, **pars):
        return MidiMessage(**self.__dict__ | pars)

    def _validate(self, path=(), names={}):
        if not hasattr(self, "type"):
            raise BankValidationError("MidiMessage missing type")
        self.type = TYPE_ALIAS[self.type]
        for par in "chan", "num", "val":
            if hasattr(self, par):
                setattr(self, par, resolve(getattr(self, par), names))

    def __repr__(self):
        return self._text


class _FlowList(list, yaml.YAMLObject):
    """
    Inline comma-separated list of YAML-embedded objects

    Compact list style with round-trip formatting
    """
    yaml_tag = "!flowlist"
    yaml_loader = BankLoader
    yaml_dumper = BankDumper
    yaml_regex = re.compile(r".*?,")

    @classmethod
    def from_yaml(cls, loader, node):
        text = loader.construct_scalar(node)
        obj = cls([yaml.load(e, Loader=BankLoader) for e in text.split(",")])
        obj._text = text
        return obj

    @classmethod
    def to_yaml(cls, dumper, data):
        return dumper.represent_scalar(cls.yaml_tag, str(data))

    def __repr__(self):
        return self._text


class _BankObject(yaml.YAMLObject):
    """
    Base class for tagged structured YAML objects inside a patch.

    Provides:
      - Storage for raw parameters as written in YAML (``_pars``)
      - Automatic deep construction of nested data
      - Validation via ``_validate``
      - Common copy/duplication support

    Subclasses define meaning, required keys, and placement
    (e.g., rules, sequences, arpeggios, etc.).
    """
    yaml_loader = BankLoader
    yaml_dumper = BankDumper

    def __init__(self, **pars):
        self._init_pars(pars)
        self._validate()

    @classmethod
    def from_yaml(cls, loader, node):
        pars = loader.construct_mapping(node, deep=True)
        obj = cls.__new__(cls)
        obj._init_pars(pars)
        return obj

    def _init_pars(self, pars):
        self.__dict__.update(pars)
        self._pars = pars

    def copy(self, **pars):
        pars["_pars"] = self._pars | pars
        return self.__class__(**self.__dict__ | pars)

    @classmethod
    def to_yaml(cls, dumper, data):
        return dumper.represent_mapping(
            cls.yaml_tag,
            data,
            flow_style=cls.yaml_flow_style
        )

    def _validate(self, path=(), names={}):
        for k, v in self:
            _walk(v, path + (k,), names)

    def __repr__(self):
        return str(self._pars)

    def __contains__(self, k):
        return k in self.__dict__

    def __iter__(self):
        return iter(
            [(k, v) for k, v in self.__dict__.items() if k[0] != "_"]
        )


class _Route:

    def __init__(self, min, max, tomin, tomax, mul=None, add=None):
        self.min = min
        self.max = max
        self.tomin = tomin
        self.tomax = tomax
        if mul == None:
            self.mul = 1 if min == max else (tomax - tomin) / (max - min)
        else:
            self.mul = mul
        if add == None:
            self.add = tomin - min * self.mul
        else:
            self.add = add

    def __iter__(self):
        tovals = range(self.tomin, self.tomax + 1)
        return iter([_Route(self.min, self.max, n, n) for n in tovals])


class MidiRule(_BankObject):
    """
    Rule describing how incoming MIDI messages are filtered,
    transformed, or mapped

    Attributes:
      type (str): Matching message type
      totype (str): Resulting message type (default: ``self.type``)
      chan (int|str): Channel range/transform (optional)
      num (int|str): Parameter range/transform (optional)
      val (int|float|str): Value range/transform (optional)
      **pars (dict): additional parameters for custom functionality
    """
    ftroute = re.compile(r"^({0})?-?({0})?=?(-?{0})?-?(-?{0})?$".format(r"[\w\d\#\.]+"))
    maroute = re.compile(r"^({0})-({0})\*(-?{0})([+-]{0})$".format(r"[\w\d\#\.]+"))
    yaml_tag = "!midirule"
    yaml_flow_style = True
    zone = "rules"
    zone_type = list

    def _validate(self, path=(), names={}):
        super()._validate(path, names)
        if not hasattr(self, "type"):
            raise BankValidationError("MidiRule missing type")
        if not hasattr(self, "totype"):
            types = self.type.split("=")
            for type in types:
                if type not in TYPE_ALIAS:
                    raise BankValidationError(
                        f"MidiRule type '{type}' not recognized"
                    )
            self.type = TYPE_ALIAS[types[0]]
            self.totype = TYPE_ALIAS[types[-1]]
        for par in "chan", "num", "val":
            spec = str(getattr(self, par, ""))
            if (m := self.ftroute.match(spec)) and any(m.groups()):
                min, max, tomin, tomax = [resolve(x, names) for x in m.groups()]
                if min == None:
                    min, max = (1, 256) if par == "chan" else (1, 127)
                elif max == None:
                    max = min
                if tomin == None:
                    tomin, tomax = min, max
                elif tomax == None:
                    tomax = tomin
                setattr(self, par, _Route(min, max, tomin, tomax))
            elif m := self.maroute.match(spec):
                min, max, mul, add = [resolve(x, names) for x in m.groups()]
                tomin = min * mul + add
                tomax = max * mul + add
                setattr(self, par, _Route(min, max, tomin, tomax, mul, add))
        if hasattr(self, "lsb"):
            self.lsb = resolve(self.lsb, names)

    def __str__(self):
        return ", ".join([f"{k}: {v}" for k, v in self._pars.items()])


class Sequence(_BankObject):
    """
    Step-sequenced event container

    Attributes:
      events (list): Parsed events grouped by pattern
      order (list): Playback order of patterns (default: ``[1]``)
      tempo (int): Beats per minute (default: ``120``)
      tdiv (int): Beat divisor (default: ``8``)
      swing (float): Timing swing ratio (default: ``0.5``)
      groove (int|list): Beat accent pattern (default: ``[1, 1]``)
    """
    yaml_tag = "!sequence"
    zone = "sequences"
    zone_type = dict

    def _init_pars(self, pars):
        super()._init_pars(pars)
        if "\n" in getattr(self, "events", ""):
            events = []
            for p in self.events.strip().split("\n\n"):
                s = [[yaml.load(e, Loader=BankLoader) for e in r.split()]
                     for r in p.splitlines()]
                events.append([list(t) for t in zip(*s)])
            self.events = events

    def _validate(self, path=(), names={}):
        super()._validate(path, names)
        if not hasattr(self, "events"):
            raise BankValidationError("Sequence objects must have events")


class Arpeggio(_BankObject):
    """
    Patterned arpeggiation definition

    Attributes:
      style (str): Pattern type
      tempo (int): Beats per minute (default: ``120``)
      tdiv (int): Beat divisor (default: ``8``)
      swing (float): Timing swing ratio (default: ``0.5``)
      groove (int|list): Beat accent pattern (default: ``[1, 1]``)
    """
    yaml_tag = "!arpeggio"
    zone = "arpeggios"
    zone_type = dict

    def _validate(self, path=(), names={}):
        super()._validate(path, names)
        if not hasattr(self, "style"):
            raise BankValidationError("Arpeggio missing style")
        if hasattr(self, "groove"):
            if isinstance(self.groove, int):
                self.groove = [self.groove, 1]


class MidiLoop(_BankObject):
    """
    Looping MIDI event generator

    Attributes:
      beats (int): Loop length in beats
      tempo (int): Beats per minute (default: ``120``)
    """
    yaml_tag = "!midiloop"
    zone = "midiloops"
    zone_type = dict

    def _validate(self, path=(), names={}):
        super()._validate(path, names)
        if not hasattr(self, "beats"):
            raise BankValidationError("MidiLoop objects must have beats")


class MidiFile(_BankObject):
    """
    MIDI file playback directive

    Attributes:
      file (str): Path to a .mid file
      tempo (int): Beats per minute (default: ``120``)
      barlength (int): Ticks per musical measure (default: ``1``)
      jumps (list[str]): List of bar jumps as [from]>[to] (default: ``[]``)
      shift (int): Channel shift amount (default: ``0``)
      mask (list[str]): Ignored message types (default: ``[]``) 
    """
    yaml_tag = "!midifile"
    zone = "midifiles"
    zone_type = dict

    def _validate(self, path=(), names={}):
        super()._validate(path, names)
        if not hasattr(self, "file"):
            raise BankValidationError("MidiFile missing file")
        if hasattr(self, "jumps"):
            if isinstance(self.jumps, str):
                self.jumps = [self.jumps]
            self.jumps = [
                [float(t) for t in s.split(">")] for s in self.jumps
            ]
        if hasattr(self, "mask"):
            self.mask = [TYPE_ALIAS[t] for t in self.mask]

class LadspaEffect(_BankObject):
    """
    LADSPA audio effect declaration

    Attributes:
      lib (str): LADSPA library basename
      plugin (str): Plugin label (default: ``''``)
      audio (list[str]): Audio port names (default: ``['Input', 'Output']``)
      vals (dict): Initial values for control ports (default: ``{}``)
      chan (list[int]): Input channel routing (default: ``[]``)
    """
    yaml_tag = "!ladspafx"
    zone = "ladspafx"
    zone_type = dict

    def _validate(self, path=(), names={}):
        super()._validate(path, names)
        if not hasattr(self, "lib"):
            raise BankValidationError("LadspaEffect missing lib")
        if hasattr(self, "chan"):
            if not isinstance(self.chan, list):
                self.chan = [self.chan]
        if hasattr(self, "audio"):
            if self.audio == "stereo":
                self.audio = "Input L", "Input R", "Output L", "Output R"
            elif self.audio == "mono":
                self.audio = "Input", "Output"


class Counter(_BankObject):
    """
    State storing object that responds to messages/rules

    Attributes:
      min/max (float): range for values
      startval (float): Initial value for counter (default: ``self.min``)
      wrap (bool): wrap values if True, else clamp to range (default: ``False``)
    """
    yaml_tag = "!counter"
    zone = "counters"
    zone_type = dict

    def _validate(self, path=(), names={}):
        super()._validate(path, names)
        if not hasattr(self, "min"):
            raise BankValidationError("Counter objects must have min")
        if not hasattr(self, "max"):
            raise BankValidationError("Counter objects must have max")
        if not hasattr(self, "startval"):
            self.startval = self.min
        if not hasattr(self, "wrap"):
            self.wrap = False

def str_presenter(dumper, data):
    if "\n" in data:
        return dumper.represent_scalar("tag:yaml.org,2002:str", str(data), style="|")
    return dumper.represent_scalar("tag:yaml.org,2002:str", data)

yaml.add_representer(str, str_presenter, Dumper=BankDumper)

for cls in (SFPreset, MidiMessage, _FlowList):
    BankLoader.add_implicit_resolver(cls.yaml_tag, cls.yaml_regex, None)
    BankDumper.add_implicit_resolver(cls.yaml_tag, cls.yaml_regex, None)

for cls in _BankObject.__subclasses__():
    path = ["patches", (dict, None), cls.zone, (cls.zone_type, None)]
    BankLoader.add_path_resolver(cls.yaml_tag, path, dict)
    BankDumper.add_path_resolver(cls.yaml_tag, path, dict)
    BankLoader.add_path_resolver(cls.yaml_tag, path[2:], dict)
    BankDumper.add_path_resolver(cls.yaml_tag, path[2:], dict)




"""
ctypes bindings and interface classes for fluidsynth
"""
from ctypes.util import find_library
from ctypes import *

FLUID_OK = 0
FLUID_FAILED = -1
FLUID_NUM_TYPE = 0
FLUID_INT_TYPE = 1
FLUID_STR_TYPE = 2
FLUID_SEQ_TIMER = 17
FLUID_SEQ_UNREGISTERING = 21
FLUID_PLAYER_TEMPO_INTERNAL = 0
FLUID_PLAYER_TEMPO_EXTERNAL_MIDI = 2
FLUID_PLAYER_PLAYING = 1
FLUID_PLAYER_DONE = 3

RULE_TYPES = "note", "ctrl", "prog", "pbend", "cpress", "kpress"
MIDI_TYPES = {
    "noteoff": 0x80, "note": 0x90, "kpress": 0xa0, "ctrl": 0xb0,
    "prog": 0xc0, "cpress": 0xd0, "pbend": 0xe0, "sysex": 0xf0,
    "clock": 0xf8, "start": 0xfa, "continue": 0xfb, "stop": 0xfc,
}
PLAYER_TYPES = "sequences", "arpeggios", "midiloops", "midifiles"
SEQ_LAG = 10

fslib = find_library("fluidsynth") or find_library("libfluidsynth-3")
if fslib is None:
    raise ImportError("Couldn't find the FluidSynth library.")
FS = CDLL(fslib)
ladspa_available = hasattr(FS, "fluid_ladspa_reset")

# type hints for ctypes as needed
FS.new_fluid_midi_event.restype = c_void_p
FS.new_fluid_event.restype = c_void_p
FS.fluid_preset_get_name.restype = c_char_p
FS.fluid_sfont_iteration_next.restype = c_void_p
FS.fluid_synth_handle_midi_event.argtypes = c_void_p, c_void_p
FS.fluid_midi_router_handle_midi_event.argtypes = c_void_p, c_void_p
fl_eventcallback = CFUNCTYPE(c_int, c_void_p, c_void_p)


class FluidMidiEvent:

    def __init__(self, e):
        b = FS.fluid_midi_event_get_type(c_void_p(e))
        self.type = {v: k for k, v in MIDI_TYPES.items()}.get(b)
        if self.type == "noteoff":
            self.type = "note"
            self.chan = FS.fluid_midi_event_get_channel(c_void_p(e)) + 1
            self.num = FS.fluid_midi_event_get_control(c_void_p(e))
            self.val = 0
        elif self.type in ("note", "kpress", "ctrl"):
            self.chan = FS.fluid_midi_event_get_channel(c_void_p(e)) + 1
            self.num = FS.fluid_midi_event_get_control(c_void_p(e))
            self.val = FS.fluid_midi_event_get_value(c_void_p(e))
        elif self.type in ("prog", "cpress", "pbend"):
            self.chan = FS.fluid_midi_event_get_channel(c_void_p(e)) + 1
            self.val = FS.fluid_midi_event_get_control(c_void_p(e))

    def __repr__(self):
        return ", ".join([f"{k}={v}" for k, v in self.__dict__.items()])


class SeqClient:

    def __init__(self, synth):
        self.synth = synth
        self.callback = CFUNCTYPE(None, c_uint, c_void_p, c_void_p, c_void_p)(
            lambda t, evt, seq, _: self.scheduler()
        )
        self.id = FS.fluid_sequencer_register_client(synth.fseq, b"", self.callback, None)
        self.ticksperbeat = 500 # default 120bpm at 1000 ticks/sec
        self.pos = 0
        self.playing = False

    def set_tempo(self, bpm):
        # default fluid_sequencer time scale is 1000 ticks per second
        self.ticksperbeat = 1000 * 60 / bpm

    def set_swing(self, swing):
        swing = min(max(0.5, swing), 0.99)
        if self.tdiv >= 8 and self.tdiv % 3:
            self.swing = [2 * swing, 2  - 2 * swing]
        else:
            self.swing = [1, 1]

    def set_groove(self, groove):
        if not isinstance(groove, list):
            groove = [groove, 1]
        self.groove = [g/max(groove) for g in groove]

    def dismiss(self):
        self.playing = False
        FS.fluid_sequencer_remove_events(
            c_void_p(self.synth.fseq), c_short(-1),
            c_short(self.id), FLUID_SEQ_TIMER
        )
        FS.fluid_sequencer_unregister_client(self.synth.fseq, self.id)


class Sequence(SeqClient):

    def __init__(self, synth, seq):
        super().__init__(synth)
        self.events = getattr(seq, "events")
        self.order = getattr(seq, "order", [1])
        self.tdiv = getattr(seq, "tdiv", 8)
        self.set_tempo(getattr(seq, "tempo", 120))
        self.set_swing(getattr(seq, "swing", 0.5))
        self.set_groove(getattr(seq, "groove", 1))

    def scheduler(self):
        if not self.playing:
            return
        if not self.events:
            return
        accent = self.groove[self.step % len(self.groove)]
        pattern = self.events[self.order[self.pos] - 1]
        for track in pattern:
            event = track[self.step % len(track)]
            if isinstance(event, str):
                continue
            if event.type == "note":
                self.synth.schedule_event(event.copy(val=event.val * accent), self.id, self.nexttick)
                dur = 0
                for i in range(len(track) - 1):
                    dur += self.swing[(self.step + i) % 2] * self.ticksperbeat * 4 / self.tdiv
                    if track[(self.step + i + 1) % len(track)] != "+":
                        break
                self.synth.schedule_event(event.copy(val=0), self.id, self.nexttick + dur)
            else:
                self.synth.schedule_event(event, self.id, self.nexttick)
        dur = self.swing[self.step % 2] * self.ticksperbeat * 4 / self.tdiv
        self.step += 1
        if self.step == max([len(track) for track in pattern]):
            self.pos = self.next
            self.step = 0
            self.next = (self.pos + 1) % len(self.order)
            if self.order[self.next] < 0:
                self.next += self.order[self.next]
        if self.order[self.pos] == 0:
            self.play(0)
        else:
            self.synth.schedule_callback(self.id, self.nexttick + dur - SEQ_LAG)
            self.nexttick += dur

    def play(self, pos=-1):
        if self.playing:
            if pos == 0:
                self.playing = False
                FS.fluid_sequencer_remove_events(
                        c_void_p(self.synth.fseq), c_short(-1),
                        c_short(self.id), FLUID_SEQ_TIMER)
            elif pos > 0:
                self.next = (int(pos) - 1) % len(self.order)
        else:
            if pos > 0:
                self.pos = (int(pos) - 1) % len(self.order)                    
            elif self.order[self.pos] == 0:
                self.pos = (self.pos + 1) % len(self.order)
            self.next = (self.pos + 1) % len(self.order)
            if self.order[self.next] < 0:
                self.next += self.order[self.next]
            self.step = 0
            self.nexttick = self.synth.currenttick
            self.playing = True
            self.scheduler()


class Arpeggio(SeqClient):

    def __init__(self, synth, arp):
        super().__init__(synth)
        self.style = getattr(arp, "style")
        self.tdiv = getattr(arp, "tdiv", 8)
        self.set_tempo(getattr(arp, "tempo", 120))
        self.set_swing(getattr(arp, "swing", 0.5))
        self.set_groove(getattr(arp, "groove", 1))
        self.keysdown = set()
        self.step = 0

    def scheduler(self):
        if not self.playing:
            return
        dur = self.swing[self.step % 2] * self.ticksperbeat * 4 / self.tdiv
        accent = self.groove[self.step % len(self.groove)]
        for event in self.notes[self.step % len(self.notes)]:
            self.synth.schedule_event(event.copy(val=event.val * accent), self.id, self.nexttick)
            self.synth.schedule_event(event.copy(val=0), self.id, self.nexttick + dur)
        self.step += 1
        self.synth.schedule_callback(self.id, self.nexttick + dur - SEQ_LAG)
        self.nexttick += dur

    def add(self, note):
        if note.val > 0:
            self.keysdown.add(note)
            nd = len(self.keysdown)
        else:
            for k in self.keysdown:
                if k.chan == note.chan and k.num == note.num:
                    self.keysdown.remove(k)
                    break
            nd = -len(self.keysdown)
        notes = list(self.keysdown)
        if self.style in ("up", "down", "both"):
            notes.sort(key=lambda n: n.num)
        if self.style == "down":
            notes.reverse()
        elif self.style == "both":
            notes += notes[-2:0:-1]
        if self.style == "chord":
            self.notes = [notes]
            if self.step == 1:
                self.nexttick = self.synth.currenttick
                dur = self.swing[0] * self.ticksperbeat * 4 / self.tdiv
                accent = self.groove[0]
                self.synth.schedule_event(note.copy(val=note.val * accent), self.id, self.nexttick)
                self.synth.schedule_event(note.copy(val=0), self.id, self.nexttick + dur)
                self.nexttick += dur
        else:
            self.notes = [[n] for n in notes]
        if nd == 1:
            if not self.playing:
                self.step = 0
                self.nexttick = self.synth.currenttick
                self.playing = True
                self.scheduler()
        elif nd == 0:
            self.playing = False
            FS.fluid_sequencer_remove_events(
                    c_void_p(self.synth.fseq), c_short(-1),
                    c_short(self.id), FLUID_SEQ_TIMER)


class MidiLoop(SeqClient):

    def __init__(self, synth, loop):
        super().__init__(synth)
        self.beats = getattr(loop, "beats")
        self.set_tempo(getattr(loop, "tempo", 120))
        self.fixedbeats = bool(self.beats)
        self.recording = False
        self.events = []
        self.layers = [[]]

    def layers_to_events(self):
        self.events = [event for layer in self.layers for event in layer]
        self.events.sort(key=lambda e: e[0])

    def scheduler(self):
        if not self.playing:
            return
        if self.pos < 0:
            # finished a loop, restart
            self.starttick += self.beats * self.ticksperbeat
            self.layers_to_events()
            self.pos = 0
        if self.events[self.pos:]:
            # schedule the current event
            b, event = self.events[self.pos]
            t = self.starttick + b * self.ticksperbeat
            self.synth.schedule_event(event, self.id, t)
            self.pos += 1
        if self.events[self.pos:]:
            # schedule a callback for the next event
            t = self.starttick + self.events[self.pos][0] * self.ticksperbeat
            self.synth.schedule_callback(self.id, t - SEQ_LAG)            
        else:
            # no more events, schedule the next loop
            self.pos = -1
            nextloop = self.starttick + self.beats * self.ticksperbeat
            self.synth.schedule_callback(self.id, nextloop) # - SEQ_LAG?

    def play(self, p=1):
        self.playing = p
        if p:
            self.starttick = self.synth.currenttick
            self.layers_to_events()
            if self.beats:
                self.pos = 0
                self.scheduler()
        else:
            if self.recording:
                self.record(0)
            FS.fluid_sequencer_remove_events(
                    c_void_p(self.synth.fseq), c_short(-1),
                    c_short(self.id), FLUID_SEQ_TIMER)

    def record(self, r=1):
        if r >= 0:
            if self.layers[-1]: # events were captured
                self.layers.append([])
                if self.beats == 0:
                    # set the loop length and start from the top
                    self.beats = (
                        self.synth.currenttick - self.starttick
                    ) / self.ticksperbeat
                    self.play()
            self.recording = bool(r)
        else:
            if self.layers[-1] == []:
                r -= 1
            self.layers = self.layers[:int(r)] + [[]]
            self.layers_to_events()
            if not self.events:
                self.play(0)
                if not self.fixedbeats:
                    self.beats = 0
            elif self.playing:
                dt = self.synth.currenttick - self.starttick
                self.pos = 0
                for event in self.events:
                    if event[0] * self.ticksperbeat > dt:
                        break
                    self.pos += 1        

    def add(self, event):
        if not self.recording:
            return
        if not self.playing:
            self.play()
        b = (self.synth.currenttick - self.starttick) / self.ticksperbeat
        if self.beats:
            b %= self.beats
        self.layers[-1].append((b, event))

    def set_tempo(self, bpm):
        ticksperbeat = 1000 * 60 / bpm
        if self.playing:
            t = self.synth.currenttick
            self.starttick = t - (t - self.starttick) * ticksperbeat / self.ticksperbeat
        self.ticksperbeat = ticksperbeat


class MidiFile:

    def __init__(self, synth, mfile):
        self.fplayer = FS.new_fluid_player(synth.fsynth)
        FS.fluid_player_add(self.fplayer, str(mfile.file).encode())
        self.jumps = getattr(mfile, "jumps", [])
        self.barlength = getattr(mfile, "barlength", 1)
        if hasattr(mfile, "tempo"):
            self.set_tempo(mfile.tempo)
        self.lasttick = 0
        self.seektick = None
        if getattr(mfile, "route", 0):
            # send midifile events to the router first (experimental)
            self.frouter_handler = fl_eventcallback(FS.fluid_midi_router_handle_midi_event)
            frouter = FS.new_fluid_midi_router(synth.st, self.frouter_handler, synth.frouter)
        else:
            # send midifile events directly to the synth (default)
            self.frouter_handler = fl_eventcallback(FS.fluid_synth_handle_midi_event)
            frouter = FS.new_fluid_midi_router(synth.st, self.frouter_handler, synth.fsynth)
        FS.fluid_midi_router_clear_rules(frouter)
        for rtype in set(RULE_TYPES) - set(getattr(mfile, "mask", [])):
            rule = FS.new_fluid_midi_router_rule()
            if hasattr(mfile, "shift"):
                FS.fluid_midi_router_rule_set_chan(rule, 0, 15, c_float(1), mfile.shift)
            FS.fluid_midi_router_add_rule(frouter, rule, RULE_TYPES.index(rtype))
        self.playback_callback = fl_eventcallback(FS.fluid_midi_router_handle_midi_event)
        FS.fluid_player_set_playback_callback(self.fplayer, self.playback_callback, frouter)
        self.tickcallback = CFUNCTYPE(None, c_void_p, c_uint)(lambda _, t: self.conduct(t))
        FS.fluid_player_set_tick_callback(self.fplayer, self.tickcallback, None)
        FS.fluid_player_seek(self.fplayer, 0) # prevent skipping first note due to ?bug

    def play(self, pos=-1):
        if FS.fluid_player_get_status(self.fplayer) != FLUID_PLAYER_PLAYING:
            if pos > 0:
                seektick = int((pos - 1) * self.barlength)
                FS.fluid_player_seek(self.fplayer, seektick)
                self.lasttick = seektick
            FS.fluid_player_play(self.fplayer)
        elif pos > 0:
            self.seektick = int((pos - 1) * self.barlength)
        elif pos == 0:
            self.seektick = None
            FS.fluid_player_stop(self.fplayer)

    def conduct(self, tick):
        if self.seektick != None:
            if tick % self.barlength < (tick - self.lasttick):
                FS.fluid_player_seek(self.fplayer, self.seektick)
                self.lasttick = self.seektick
                self.seektick = None
        elif self.lasttick < tick:
            for frombar, tobar in self.jumps:
                if self.lasttick < frombar * self.barlength <= tick:
                    totick = int((tobar - 1) * self.barlength)
                    if tobar == 0:
                        self.play(0)
                        totick = 0
                    FS.fluid_player_seek(self.fplayer, totick)
                    self.lasttick = totick
                    break
            else:
                self.lasttick = tick

    def set_tempo(self, bpm=0):
        if bpm:
            usec = int(60000000.0 / bpm) # usec per quarter note (MIDI standard)
            FS.fluid_player_set_tempo(self.fplayer, FLUID_PLAYER_TEMPO_EXTERNAL_MIDI, c_double(usec))
        else:
            FS.fluid_player_set_tempo(self.fplayer, FLUID_PLAYER_TEMPO_INTERNAL, c_double(1.0))

    def dismiss(self):
        FS.fluid_player_stop(self.fplayer)
        FS.delete_fluid_player(self.fplayer)


class LadspaEffect:
    
    def __init__(self, synth, name, fx):
        self.ladspa = synth.ladspa
        lib = str(fx.lib).encode()
        plugin = getattr(fx, "plugin", "").encode() or None
        chan = getattr(fx, "chan", [])
        audio = getattr(fx, "audio", ("Input", "Output"))
        portvals = getattr(fx, "vals", {})
        groups = set([(n - 1) % len(synth.port_mapping) for n in chan])
        aports = [port.encode() for port in audio]
        stereo = True if len(aports) == 4 else False
        self.fxunits = []
        self.links = {}
        def addfxunit():
            fxunit = f"{name}{len(self.fxunits)}".encode()
            if FS.fluid_ladspa_add_effect(self.ladspa, fxunit, lib, plugin) != FLUID_OK:
                return False
            if FS.fluid_ladspa_effect_can_mix(self.ladspa, fxunit):
                FS.fluid_ladspa_effect_set_mix(self.ladspa, fxunit, 1, c_float(1.0))
            for port, val in portvals.items():
                FS.fluid_ladspa_effect_set_control(self.ladspa, fxunit, port.encode(), c_float(val))
            self.fxunits.append(fxunit)
            return True
        for group, (hostports, outports) in enumerate(synth.port_mapping):
            if groups and group not in groups:
                continue
            if stereo:
                if addfxunit():
                    self.links[hostports] = self.fxunits[-1:] * 2, aports[0:2], aports[2:4]
            else:
                if addfxunit() and addfxunit():
                    self.links[hostports] = self.fxunits[-2:], aports[0:1] * 2, aports[1:2] * 2

    def link(self, hostports, inputs, outputs):
        for fxunit, fxin, fxout, inp, outp in zip(*self.links[hostports], inputs, outputs):
            FS.fluid_ladspa_effect_link(self.ladspa, fxunit, fxin, inp.encode())
            FS.fluid_ladspa_effect_link(self.ladspa, fxunit, fxout, outp.encode())

    def setcontrol(self, port, val):
        for fxunit in self.fxunits:
            FS.fluid_ladspa_effect_set_control(self.ladspa, fxunit, port.encode(), c_float(val))


class SoundFont:
    """An iterable soundfont container"""

    def __init__(self, fsfont, sfid):
        self.id = sfid
        self._presets = {}
        FS.fluid_sfont_iteration_start(fsfont)
        while True:
            p = FS.fluid_sfont_iteration_next(fsfont)
            if p == None: break
            bank = FS.fluid_preset_get_banknum(p)
            prog = FS.fluid_preset_get_num(p)
            name = FS.fluid_preset_get_name(p).decode()
            self._presets[bank, prog] = name

    def __getitem__(self, p):
        return self._presets[p] if p in self._presets else ""

    def index(self, p):
        return list(self._presets).index(p)

    def items(self):
        return self._presets.items()

    def __iter__(self):
        return iter(self._presets)

    def __len__(self):
        return len(self._presets)


class Synth:

    def __init__(self, fluidsettings={}, logfunc=None, midi_handler=None):
        self.st = FS.new_fluid_settings()
        for name, val in fluidsettings.items():
            self[name] = val
        if logfunc:
            self.logfunc = CFUNCTYPE(None, c_int, c_char_p, c_void_p)(
                lambda lev, msg, _: logfunc(lev, msg)
            )
            for lev in range(5):
                FS.fluid_set_log_function(lev, self.logfunc)
        self.fsynth = FS.new_fluid_synth(self.st)
        FS.new_fluid_audio_driver(self.st, self.fsynth)
        self.frouter_handler = fl_eventcallback(FS.fluid_synth_handle_midi_event)
        self.frouter = FS.new_fluid_midi_router(self.st, self.frouter_handler, self.fsynth)
        if midi_handler:
            self.fdriver_handler = fl_eventcallback(
                lambda _, e: midi_handler(FluidMidiEvent(e)) or FLUID_OK
            )
        else:
            self.fdriver_handler = fl_eventcallback(FS.fluid_midi_router_handle_midi_event)
        FS.new_fluid_midi_driver(self.st, self.fdriver_handler, self.frouter)
        self.fseq = FS.new_fluid_sequencer2(0)
        self.id = FS.fluid_sequencer_register_fluidsynth(self.fseq, self.fsynth)
        self.players = {ptype: {} for ptype in PLAYER_TYPES}
        self.ladspafx = {}
        if ladspa_available:
            if self["synth.audio-groups"] == 1:
                hostports = outports = [("Main:L", "Main:R")]
            else:
                hostports = [(f"Main:L{i}", f"Main:R{i}") for i in range(1, self["synth.audio-groups"] + 1)]
                outports = hostports[0:self["synth.audio-channels"]] * self["synth.audio-groups"]
            self.port_mapping = list(zip(hostports, outports))
            self.ladspa = FS.fluid_synth_get_ladspa_fx(self.fsynth)

    @property
    def currenttick(self):
        return FS.fluid_sequencer_get_tick(self.fseq)

    def reset(self):
        for ptype in PLAYER_TYPES:
            for name in list(self.players[ptype]):
                self.player_remove(ptype, name)
        self.fxchain_clear()
        FS.fluid_synth_system_reset(self.fsynth)

    def __getitem__(self, name):
        stype = FS.fluid_settings_get_type(self.st, name.encode())
        if stype == FLUID_STR_TYPE:
            strval = create_string_buffer(32)
            if FS.fluid_settings_copystr(self.st, name.encode(), strval, 32) == FLUID_OK:
                return strval.value.decode()
        elif stype == FLUID_INT_TYPE:
            val = c_int()
            if FS.fluid_settings_getint(self.st, name.encode(), byref(val)) == FLUID_OK:
                return val.value
        elif stype == FLUID_NUM_TYPE:
            num = c_double()
            if FS.fluid_settings_getnum(self.st, name.encode(), byref(num)) == FLUID_OK:
                return round(num.value, 6)
        return None

    def __setitem__(self, name, val):
        stype = FS.fluid_settings_get_type(self.st, name.encode())
        if stype == FLUID_STR_TYPE:
            FS.fluid_settings_setstr(self.st, name.encode(), str(val).encode())
        elif stype == FLUID_INT_TYPE:
            FS.fluid_settings_setint(self.st, name.encode(), int(val))
        elif stype == FLUID_NUM_TYPE:
            FS.fluid_settings_setnum(self.st, name.encode(), c_double(float(val)))

    def load_soundfont(self, path):
        sfid = FS.fluid_synth_sfload(self.fsynth, str(path).encode(), False)
        if sfid == FLUID_FAILED:
            raise OSError(f"Unable to load {file}")
        fsfont = FS.fluid_synth_get_sfont_by_id(self.fsynth, sfid)
        return SoundFont(fsfont, sfid)

    def unload_soundfont(self, sfont):
        FS.fluid_synth_sfunload(self.fsynth, sfont.id, False)

    def program_select(self, chan, sfont, bank, prog):
        x = FS.fluid_synth_program_select(self.fsynth, chan - 1, sfont.id, bank, prog)
        return True if x == FLUID_OK else False

    def program_unset(self, chan):
        FS.fluid_synth_unset_program(self.fsynth, chan - 1)

    def program_info(self, chan):
        i = c_int()
        bank = c_int()
        prog = c_int()
        FS.fluid_synth_get_program(self.fsynth, chan - 1, byref(i), byref(bank), byref(prog))
        return i.value, bank.value, prog.value

    def get_cc(self, chan, num):
        val = c_int()
        FS.fluid_synth_get_cc(self.fsynth, chan - 1, num, byref(val))
        return val.value

    def router_clear(self):
        FS.fluid_midi_router_clear_rules(self.frouter)

    def router_default(self):
        FS.fluid_midi_router_set_default_rules(self.frouter)

    def router_addrule(self, rule):
        frule = FS.new_fluid_midi_router_rule()
        if chan := getattr(rule, "chan", None):
            FS.fluid_midi_router_rule_set_chan(
                frule, int(chan.min - 1), int(chan.max - 1),
                c_float(chan.mul), int(chan.add + chan.mul - 1)
            )
        if rule.type in ("prog", "cpress", "pbend"):
            if val := getattr(rule, "val", None):
                FS.fluid_midi_router_rule_set_param1(
                    frule, int(val.min), int(val.max),
                    c_float(val.mul), int(val.add)
                )
        else:
            if num := getattr(rule, "num", None):
                FS.fluid_midi_router_rule_set_param1(
                    frule, int(num.min), int(num.max),
                    c_float(num.mul), int(num.add)
                )
            if val := getattr(rule, "val", None):
                FS.fluid_midi_router_rule_set_param2(
                    frule, int(val.min), int(val.max),
                    c_float(val.mul), int(val.add)
                )
        FS.fluid_midi_router_add_rule(self.frouter, frule, RULE_TYPES.index(rule.type))

    def send_midievent(self, event, route=False):
        fmevent = FS.new_fluid_midi_event()
        if event.type == "sysex":
            syxdata = (c_int * len(event.val))(*event.val)
            FS.fluid_midi_event_set_sysex(c_void_p(fmevent), syxdata, sizeof(syxdata), True)
        else:
            FS.fluid_midi_event_set_type(c_void_p(fmevent), MIDI_TYPES.get(event.type))
            if event.type in ("prog", "cpress", "pbend"):
                FS.fluid_midi_event_set_channel(c_void_p(fmevent), int(event.chan - 1))
                FS.fluid_midi_event_set_control(c_void_p(fmevent), int(event.val))
            elif event.type in ("note", "kpress", "ctrl"):
                FS.fluid_midi_event_set_channel(c_void_p(fmevent), int(event.chan - 1))
                FS.fluid_midi_event_set_control(c_void_p(fmevent), int(event.num))
                FS.fluid_midi_event_set_value(c_void_p(fmevent), int(event.val))
        if route:
            FS.fluid_midi_router_handle_midi_event(self.frouter, fmevent)
        else:
            FS.fluid_synth_handle_midi_event(self.fsynth, fmevent)

    def schedule_event(self, event, id=-1, time=None):
        if not hasattr(event, "chan"):
            return
        fevent = FS.new_fluid_event()
        FS.fluid_event_set_source(c_void_p(fevent), c_short(id))
        FS.fluid_event_set_dest(c_void_p(fevent), c_short(self.id))
        chan, val = int(event.chan - 1), int(event.val)
        if event.type in ("note", "ctrl", "kpress"):
            num = int(event.num)
        if event.type == "note":
            FS.fluid_event_noteon(c_void_p(fevent), chan, num, val)
        elif event.type == "ctrl":
            FS.fluid_event_control_change(c_void_p(fevent), chan, num, val)
        elif event.type == "prog":
            FS.fluid_event_program_change(c_void_p(fevent), chan, val)
        elif event.type == "pbend":
            FS.fluid_event_pitch_bend(c_void_p(fevent), chan, val)
        elif event.type == "cpress":
            FS.fluid_event_channel_pressure(c_void_p(fevent), chan, val)
        elif event.type == "kpress":
            FS.fluid_event_key_pressure(c_void_p(fevent), chan, num, val)
        if time == None:
            time = self.currenttick
        FS.fluid_sequencer_send_at(
                c_void_p(self.fseq), c_void_p(fevent),
                c_uint(int(time)), 1)
        FS.delete_fluid_event(c_void_p(fevent))

    def schedule_callback(self, id, time):
        cb = FS.new_fluid_event()
        FS.fluid_event_set_source(c_void_p(cb), c_short(-1))
        FS.fluid_event_set_dest(c_void_p(cb), c_short(id))
        FS.fluid_event_timer(c_void_p(cb), c_void_p())
        FS.fluid_sequencer_send_at(
                c_void_p(self.fseq), c_void_p(cb),
                c_uint(int(time)), 1)
        FS.delete_fluid_event(c_void_p(cb))

    def player_add(self, ptype, name, player):
        if name not in self.players[ptype]:
            if ptype == "sequences":
                self.players[ptype][name] = Sequence(self, player)
            elif ptype == "arpeggios":
                self.players[ptype][name] = Arpeggio(self, player)
            elif ptype == "midiloops":
                self.players[ptype][name] = MidiLoop(self, player)
            elif ptype == "midifiles":
                self.players[ptype][name] = MidiFile(self, player)
    
    def player_remove(self, ptype, name):
        if name in self.players[ptype]:
            self.players[ptype][name].dismiss()
            del self.players[ptype][name]

    def fxchain_clear(self):
        FS.fluid_ladspa_reset(self.ladspa)
        self.ladspafx = {}

    def fxchain_add(self, name, fx):
        if name not in self.ladspafx:
            self.ladspafx[name] = LadspaEffect(self, name, fx)

    def fxchain_connect(self):
        b = -1
        for hostports, outports in self.port_mapping:
            effects = [e for e in self.ladspafx.values() if hostports in e.links]
            if not effects: continue
            lastports = hostports
            for effect in effects[0:-1]:
                b += 2
                buffers = (f"buffer{b}", f"buffer{b + 1}")
                FS.fluid_ladspa_add_buffer(self.ladspa, buffers[0].encode())
                FS.fluid_ladspa_add_buffer(self.ladspa, buffers[1].encode())
                effect.link(hostports, lastports, buffers)
                lastports = buffers
            effects[-1].link(hostports, lastports, outports)
        FS.fluid_ladspa_activate(self.ladspa)

    if not ladspa_available:
        def fxchain_clear(self): pass
        def fxchain_add(self, _, __): pass
        def fxchain_connect(self): pass




#!/usr/bin/env python3
"""
Adding and removing patches at runtime

Normally banks are created and loaded as YAML. This example shows how
to modify a bank within a program.
"""

from fluidpatcher import FluidPatcher, SFPreset, MidiRule

# the MIDI channel of notes from your controller
MIDI_CHANNEL = 1

# setup, suppress fluidsynth logging
fp = FluidPatcher(fluidlog=-1)

# load a soundfont
soundfont = fp.open_soundfont("test.sf2")
print(f"Loaded soundfont: {soundfont.file}")

# create a patch
fp.bank.patch["Piano"] = {
    MIDI_CHANNEL: SFPreset(soundfont.file, 0, 4)
}

# set a rule at the root level so notes go to the synth
fp.bank.root["rules"] = [
    MidiRule(type="note", chan=MIDI_CHANNEL)
]

current_patch = ""
# main loop
while True:
    print("Patches:")
    for i, name in enumerate(fp.bank.patches, start=1):
        print(">" if name == current_patch else " ", end="")
        print(f"{i:>5}) {name}")

    cmd = input(
        "# = select patch | a = add patch | c = change preset | "
        "d = delete patch | q = quit: "
    ).lower()

    if cmd.isdigit():
        if 0 < int(cmd) <= len(fp.bank.patches):
            current_patch = fp.bank.patches[int(cmd) - 1]
            fp.apply_patch(current_patch)
        else:
            print("Patch number out of range")

    elif cmd == "a":
        name = input("Enter new patch name: ")
        if name == "":
            continue
        elif name in fp.bank.patches:
            print(f"Patch {name} already exists")
            continue
        # add the patch
        fp.bank.patch[name] = {}

        # list available presets
        print("preset bank:prog name")
        print("====== ========= ====")

        for i, (bank, prog) in enumerate(soundfont, start=1):
            print(f"{i:>6}  {bank:03}:{prog:03}  {soundfont[bank, prog]}")

        p = input("Enter preset #: ")
        if p.isdigit():
            i = int(p) - 1
            if 0 <= i < len(soundfont):
                bank, prog = list(soundfont)[i]
                # modify the patch
                fp.bank.patch[name][MIDI_CHANNEL] = SFPreset(
                    soundfont.file, bank, prog
                )
            else:
                print("Preset number out of range")

    elif cmd == "d":
        name = input("Enter name of patch to delete: ")
        if name in fp.bank.patches:
            # delete the patch
            del fp.bank.patch[name]
        else:
            print(f"Patch name '{name}' not found")

    elif cmd == "q":
        break




"""Performance-oriented patch management for FluidSynth.

FluidPatcher provides a high-level interface to the FluidSynth
software synthesizer, organized around *patches* stored in YAML
bank files. Each patch can define instruments, controller values,
LADSPA effects, MIDI sequences, router rules, and more.

Key components:
  - patcher.py - user-facing API for loading banks, applying patches
  - bankfiles.py – YAML extensions and helpers for parsing banks
  - config.py - configuration loading and initialization
  - router.py – live MIDI routing and rule processing
  - pfluidsynth.py – ctypes bindings and custom implementations
    as lightweight wrappers around FluidSynth objects

Typical usage:

    fp = FluidPatcher()
    fp.load_bank("mybank.yaml")
    fp.apply_patch("Warm Pad")

Requirements:
  - PyYAML
  - libfluidsynth
"""

from importlib.metadata import version, PackageNotFoundError

from .bankfiles import MidiMessage, MidiRule, SFPreset
from .config import CONFIG, save_config
from .patcher import FluidPatcher
from .pfluidsynth import FluidMidiEvent


__all__ = ["MidiMessage", "MidiRule", "SFPreset", "CONFIG",
           "save_config", "FluidPatcher", "FluidMidiEvent"]

try:
    __version__ = version("fluidpatcher")
except PackageNotFoundError:
    __version__ = "0.0.0-dev"




"""
Python handler for fluidsynth midi events
with extensible custom router rules
"""

from .pfluidsynth import PLAYER_TYPES


class RouterEvent:

    def __init__(self, event, rule=None):
        self.__dict__.update(event.__dict__)
        self.rule = rule
        
    def copy(self, **pars):
        e = RouterEvent(self)
        e.__dict__.update(pars)
        return e


class RouterRule:

    def __init__(self, rule):
        self.__dict__.update(rule.__dict__)

    def applies(self, event):
        if self.type != event.type:
            return False
        for par in ("chan", "num", "val"):
            if hasattr(self, par) and hasattr(event, par):
                r = getattr(self, par)
                if r.max < r.min:
                    if r.max < getattr(event, par) < r.min:
                        return False
                elif not (r.min <= getattr(event, par) <= r.max):
                    return False
        return True

    def apply(self, event):
        newevent = RouterEvent(event, self)
        newevent.type = self.totype
        if hasattr(event, "chan"):
            if hasattr(self, "chan"):
                newevent.chan = int(event.chan * self.chan.mul + self.chan.add)
            if hasattr(self, "val"):
                if hasattr(self, "log"):
                    b = 10 ** self.log
                    x = (event.val - self.val.min) / (self.val.max - self.val.min)
                    newevent.val = self.val.tomin + (self.val.tomax - self.val.tomin) * (b ** x - 1) / (b - 1)
                else:
                    newevent.val = event.val * self.val.mul + self.val.add
                if hasattr(self, "lsb"):
                    newevent.lsbval = newevent.val % 127
                    newevent.val //= 127
            if hasattr(self, "num"):
                if self.totype in ("note", "ctrl", "kpress"):
                    if hasattr(event, "num"):
                        newevent.num = round(event.num * self.num.mul + self.num.add)
                    else:
                        newevent.num = self.num.min
        else:
            if hasattr(self, "chan"):
                newevent.chan = self.chan.min
            if hasattr(self, "num"):
                newevent.num = self.num.min
            if hasattr(self, "val"):
                newevent.val = self.val.min
            elif event.type == "clock":
                newevent.val = 1/24
            elif event.type in ("start", "continue"):
                newevent.val = -1
            elif event.type == "stop":
                newevent.val = 0
        return newevent


class FluidRule:

    def __init__(self, rule):
        self.type = rule.type
        for par in "chan", "num", "val":
            if route := getattr(rule, par, None):
                setattr(self, par, route)


class Router:

    def __init__(self, fluid_default=False, fluid_router=False):
        self.fluid_default = fluid_default
        self.fluid_router = fluid_router
        self.rules = []
        self.fluidrules = []
        self.counters = {}
        self.synth = None
        self.callback = lambda event: None
        self.clocks = [0, 0]

    def reset(self):
        self.rules = []
        self.fluidrules = []
        if self.fluid_router:
            self.synth.router_clear()
            if self.fluid_default:
                self.synth.router_default()

    def add(self, rule):
        if (
            self.fluid_router
            and rule.type == rule.totype
            and rule.type in ("note", "kpress", "ctrl", "prog", "cpress", "pbend")
            and not set(rule.__dict__) - {"type", "totype", "chan", "num", "val", "_pars"}
           ):
            if hasattr(rule, "chan"):
                for tochan in rule.chan:
                    self.fluidrules.append(FluidRule(rule.copy(chan=tochan)))
            else:
                self.fluidrules.append(FluidRule(rule))
            self.synth.router_clear()
            if self.fluid_default:
                self.synth_router_default()
            # add rules in reverse order because fluidsynth handles them LIFO-style
            for rule in self.fluidrules[::-1]:
                self.synth.router_addrule(rule)
        else:
            if hasattr(rule, "chan"):
                for tochan in rule.chan:
                    self.rules.append(RouterRule(rule.copy(chan=tochan)))
            else:
                self.rules.append(RouterRule(rule))

    def find_players(self, name):
        return [self.synth.players[ptype][name]
                for ptype in PLAYER_TYPES
                if name in self.synth.players[ptype]]

    def handle_midi(self, event):
        if self.fluid_router:
            # let fluidsynth route voice events
            self.synth.send_midievent(event, route=True)
        self.callback(event) # forward it to the callback
        t = self.synth.currenttick
        dt = 0
        for rule in [r for r in self.rules if r.applies(event)]:
            newevent = rule.apply(event)
            if hasattr(rule, "counter"):
                if rule.counter in self.counters:
                    c = self.counters[rule.counter]
                    c.val += newevent.val
                    if c.val > c.max:
                        c.val = c.min if c.wrap else c.max
                    elif c.val < c.min:
                        c.val = c.max if c.wrap else c.min
                    newevent.val = c.val
            if hasattr(rule, "lsb"):
                lsbevent = RouterEvent(newevent, rule)
                lsbevent.num, lsbevent.val = rule.lsb, newevent.lsbval
                self.synth.send_midievent(lsbevent)
            if hasattr(rule, "fluidsetting"):
                self.synth[rule.fluidsetting] = newevent.val
            if hasattr(rule, "play"):
                for player in self.find_players(rule.play):
                    player.play(newevent.val)
            if hasattr(rule, "tempo"):
                for player in self.find_players(rule.tempo):
                    player.set_tempo(newevent.val)
            if hasattr(rule, "tap"):
                for player in self.find_players(rule.tap):
                    dt, dt2 = t - self.clocks[0], self.clocks[0] - self.clocks[1]
                    if dt2/dt > 0.5: # wait for three taps of similar spacing
                        bpm = 1000 * 60 * newevent.val / dt
                        player.set_tempo(bpm)
            if hasattr(rule, "record"):
                for player in self.find_players(rule.record):
                    if hasattr(player, "record"):
                        player.record(newevent.val)
            if hasattr(rule, "arpeggio"):
                for player in self.find_players(rule.arpeggio):
                    if hasattr(player, "add"):
                        player.add(newevent.copy())
                        newevent.val = 0
            if hasattr(rule, "loop"):
                for player in self.find_players(rule.loop):
                    if hasattr(player, "add"):
                        player.add(newevent.copy())
            if hasattr(rule, "swing"):
                for player in self.find_players(rule.swing):
                    if hasattr(player, "set_swing"):
                        player.set_swing(newevent.val)
            if hasattr(rule, "groove"):
                for player in self.find_players(rule.groove):
                    if hasattr(player, "set_groove"):
                        player.set_groove(newevent.val)
            if hasattr(rule, "fx"):
                fx, port = rule.fx.split(">")
                if fx in self.synth.ladspafx:
                    self.synth.ladspafx[fx].setcontrol(port, newevent.val)
            self.synth.send_midievent(newevent) # send routed event to synth
            self.callback(newevent) # forward the routed event for user handling
        if dt > 0:
            self.clocks = t, self.clocks[0]




"""
Configuration loading and initialization for fluidpatcher.

This module defines where fluidpatcher stores its configuration data
(~/.config/fluidpatcher/... by default), loads settings on import,
creates missing directories, and populates them with bundled defaults
when needed.

At runtime:
- CONFIG holds the merged configuration mapping
- CONFIG_PATH stores the config file path
- PATCHCORD resolves to a LADSPA plugin .so suitable for this machine,
  or None if no build is available
"""
import importlib.resources as res
import os
from pathlib import Path
import platform
import shutil

import yaml

from .bankfiles import LadspaEffect


def load_config():
    cfg = yaml.safe_load(CONFIG_PATH.read_text())
    for key, val in list(cfg.items()):
        if key.endswith("_path") and val is not None:
            cfg[key] = Path(val)
    cfg.setdefault("banks_path", CONFIG_PATH.parent / "banks")
    cfg.setdefault("sounds_path", cfg["banks_path"].parent / "sounds")
    cfg.setdefault("midi_path", cfg["banks_path"].parent / "midi")
    cfg.setdefault(
        "ladspa_path",
        Path(os.getenv("LADSPA_PATH", "/usr/lib/ladspa"))
    )
    return cfg


def save_config():
    """
    Writes CONFIG to file.
    """
    cfg_posix = {
        k: v.as_posix() if isinstance(v, Path) else v
        for k, v in CONFIG.items()
    }
    CONFIG_PATH.write_text(
        yaml.safe_dump(cfg_posix, sort_keys=False)
    )


CONFIG_PATH = Path(os.getenv(
    "FLUIDPATCHER_CONFIG",
    "~/.config/fluidpatcher/fluidpatcherconf.yaml"
)).expanduser()

if not CONFIG_PATH.exists():
    CONFIG_PATH.parent.mkdir(parents=True, exist_ok=True)
    shutil.copy(res.files("fluidpatcher.data") / "fluidpatcherconf.yaml", CONFIG_PATH)

CONFIG = load_config()
if not CONFIG["banks_path"].exists():
    shutil.copytree(res.files("fluidpatcher.data") / "banks", CONFIG["banks_path"])
if not CONFIG["sounds_path"].exists():
    shutil.copytree(res.files("fluidpatcher.data") / "sounds", CONFIG["sounds_path"])
if not CONFIG["midi_path"].exists():
    shutil.copytree(res.files("fluidpatcher.data") / "midi", CONFIG["midi_path"])

patchcord = res.files("fluidpatcher._ladspa") / "patchcord.so" 
system = platform.system().lower()
arch = platform.machine()
prebuilt_path = res.files("fluidpatcher._ladspa") / f"prebuilt/{system}-{arch}/patchcord.so"

if patchcord.exists():
    PATCHCORD = {"_patchcord": LadspaEffect(lib=patchcord)}
elif prebuilt_path.exists():
    PATCHCORD = {"_patchcord": LadspaEffect(lib=prebuilt_path)}
else:
    PATCHCORD = {}
    CONFIG["fluidsettings"]["synth.audio-groups"] = 1




#!/usr/bin/env python3
"""
Minimal FluidPatcher example.

Loads a bank and lets you step through patches.
Play notes from any connected MIDI controller.
"""

from fluidpatcher import FluidPatcher, CONFIG

# setup
fp = FluidPatcher()

# load a bank
fp.load_bank("testbank.yaml")

print(f"Loaded bank: testbank.yaml")
print("Patches:")
for i, name in enumerate(fp.bank.patches, start=1):
    print(f"{i:>5}) {name}")
    
# select the first patch
patch_index = 0
fp.apply_patch(fp.bank.patches[patch_index])

# main loop
while True:
    print(
        f"Applied patch: {fp.bank.patches[patch_index]} "
        f"({patch_index + 1}/{len(fp.bank.patches)})"
    )

    cmd = input("Enter = next patch | q = quit: ").lower()

    if cmd == "q":
        break
    else:
        patch_index = (patch_index + 1) % len(fp.bank.patches)
        patch_name = fp.bank.patches[patch_index]
        fp.apply_patch(patch_name)




import pathlib
import platform
import subprocess
import sysconfig

from setuptools import setup
from setuptools.command.build_py import build_py


class build_py_with_patchcord(build_py):
    def run(self):
        super().run()
        self.build_patchcord()

    def build_patchcord(self):
        system = platform.system().lower()
        if system != "linux":
            print("Skipping patchcord build (non-Linux platform)")
            return

        src = pathlib.Path("src/fluidpatcher/_ladspa/patchcord.c")

        if not src.exists():
            print("patchcord source missing, skipping build")
            return

        target = pathlib.Path(self.build_lib) / "fluidpatcher/_ladspa/patchcord.so"
        target.parent.mkdir(parents=True, exist_ok=True)

        cc = sysconfig.get_config_var("CC") or "cc"
        cflags = sysconfig.get_config_var("CFLAGS") or ""

        cmd = [
            cc,
            "-shared",
            "-fPIC",
            *cflags.split(),
            str(src),
            "-o",
            str(target),
        ]

        try:
            print("Compiling patchcord LADSPA plugin...")
            subprocess.check_call(cmd)
            print("patchcord build complete")
        except Exception as e:
            print("patchcord build failed:", e)
            print("Continuing without patchcord support")


setup(
    cmdclass={
        "build_py": build_py_with_patchcord,
    },
)




"""
High-level API for live patch control.

This module coordinates interaction between the synth
engine and Bank files. It handles:

  - loading and saving of banks
  - applying patch parameters (programs, controllers, volumes, etc.)
  - direct interaction with the synth and router
  - optional notification hooks for UI or CLI environments
"""

from pathlib import Path

from yaml import safe_load, safe_dump

from .bankfiles import Bank, SFPreset, MidiMessage
from .bankfiles import BankValidationError
from .config import CONFIG, PATCHCORD
from .pfluidsynth import Synth, PLAYER_TYPES
from .router import Router


CC_DEFAULTS = [0] * 120
CC_DEFAULTS[0] = -1             # bank select
CC_DEFAULTS[7] = 100            # volume
CC_DEFAULTS[8] = 64             # balance
CC_DEFAULTS[10] = 64            # pan
CC_DEFAULTS[11] = 127           # expression
CC_DEFAULTS[32] = -1            # bank select LSB
CC_DEFAULTS[43] = 127           # expression LSB
CC_DEFAULTS[70:80] = [64] * 10  # sound controls
CC_DEFAULTS[84] = 255           # portamento control
CC_DEFAULTS[96:102] = [-1] * 6  # RPN/NRPN controls

SYNTH_DEFAULTS = {"synth.chorus.active": 1, "synth.reverb.active": 1,
                  "synth.chorus.depth": 8.0, "synth.chorus.level": 2.0,
                  "synth.chorus.nr": 3, "synth.chorus.speed": 0.3,
                  "synth.reverb.damp": 0.0, "synth.reverb.level": 0.9,
                  "synth.reverb.room-size": 0.2, "synth.reverb.width": 0.5,
                  "synth.gain": 0.2}


class FluidPatcher:
    """
    High-level controller FluidSynth. Applies descriptive YAML-based
    patches and manages synth state.

    Attributes:
      bank (Bank):
        Parsed representation of the currently loaded bank.

      soundfonts (dict[path, SoundFont]):
        Mapping of loaded soundfonts, keyed by file path.                  
    """

    def __init__(self, fluidsettings={}, fluidlog=None):
        """
        Create a FluidPatcher and start FluidSynth.

        Args:
          fluidsettings (dict):
            Additional FluidSynth settings that override defaults.

          fluidlog (callable | -1 | None):
            Callback accepting (level, message) or -1 to suppress logs.
        """
        self.bank = Bank("patches: {}")
        self._sfonts = {}
        self._router = Router(fluid_default=False, fluid_router=False)
        if fluidlog == -1:
            fluidlog = lambda lev, msg: None
        self._synth = Synth(
            fluidsettings=CONFIG["fluidsettings"] | fluidsettings,
            logfunc=fluidlog,
            midi_handler=self._router.handle_midi,
        )
        self._router.synth = self._synth

    @property
    def soundfonts(self):
        """dict[path, SoundFont]: A snapshot of the loaded soundfonts."""
        return self._sfonts

    def open_soundfont(self, path):
        """
        Load a soundfont and enumerate its presets.

        Args:
          path (str|Path):
            Filename relative to CONFIG["sounds_path"], or absolute.

        Returns:
          (SoundFont): iterable of presets, indexable by (bank, prog).
        """
        path = CONFIG["sounds_path"] / path
        if path.is_relative_to(CONFIG["sounds_path"]):
            sf = path.relative_to(CONFIG["sounds_path"]).as_posix()
        else:
            sf = path.as_posix()
        if sf not in self._sfonts:
            self._sfonts[sf] = self._synth.load_soundfont(path)
            self._sfonts[sf].file = sf
        return self._sfonts[sf]

    def load_bank(self, bankfile="", raw=""):
        """
        Load a bank from a YAML file or raw text.

        Args:
          bankfile (str|Path):
            Filename relative to CONFIG["banks_path"], or absolute.

          raw (str):
            YAML text to load directly, bypassing disk I/O.
        """
        def read_bank(files, raw="", indent=0):
            text = ""
            if files:
                try:
                    raw = (CONFIG["banks_path"] / files[-1]).read_text()
                except FileNotFoundError as e:
                    raise BankValidationError(
                        f"No such file {files[-1]}"
                    )
            for line in raw.splitlines(keepends=True):
                if "#include" in line:
                    i = line.index("#include")
                    f = line[i + 9:].rstrip()
                    if f not in files:
                        if line[:i].isspace():
                            line = read_bank(files + [f], indent=i)
                        else:
                            line = line[:i] + read_bank(files + [f])
                text += " " * indent + line
            return text
        self.bank = Bank(read_bank(
            [bankfile] if bankfile else [], raw
        ))
        self._synth.reset()
        for zone in self.bank:
            for midi in zone.get("midifiles", {}).values():
                midi.file = CONFIG["midi_path"] / midi.file
            for fx in zone.get("ladspafx", {}).values():
                fx.lib = CONFIG["ladspa_path"] / fx.lib
        init = self.bank.root.get("init", {})
        for name, val in (SYNTH_DEFAULTS | init.get("fluidsettings", {})).items():
            self._synth[name] = val
        for msg in init.get("messages", []):
            self.send_midimessage(msg)

    def save_bank(self, bankfile, raw=""):
        """
        Write the current bank contents to disk.

        Args:
          bankfile (str|Path):
            Output filename relative to CONFIG["banks_path"], or absolute

          raw (str):
            Exact YAML text to store.
        """
        if raw:
            self.bank = Bank(raw)
        else:
            raw = self.bank.dump()
        (CONFIG["banks_path"] / bankfile).write_text(raw)

    def apply_patch(self, patch):
        """
        Apply a named patch from the loaded bank.

        Args:
          patch (str): The patch name to apply.
        """
        # load all needed soundfonts at once to speed up patches
        # free memory of unneeded soundfonts
        for sf in set(self._sfonts) - self.bank.soundfonts:
            self._synth.unload_soundfont(self._sfonts[sf])
            del self._sfonts[sf]
        for sf in self.bank.soundfonts - set(self._sfonts):
            self.open_soundfont(sf)
        # select presets
        for chan in range(1, self._synth["synth.midi-channels"] + 1):
            if p := self.bank[patch][chan]:
                self._synth.program_select(chan, self._sfonts[p.file], p.bank, p.prog)
            else:
                self._synth.program_unset(chan)
        # fluidsettings
        for name, val in self.bank[patch]["fluidsettings"].items():
            self._synth[name] = val
        # players (e.g. sequences, arpeggios, midiloops, midifiles)
        for ptype in PLAYER_TYPES:
            for name in list(self._synth.players[ptype]):
                if name not in self.bank[patch][ptype]:
                    self._synth.player_remove(ptype, name)
            for name, player in self.bank[patch][ptype].items():
                if name not in self._synth.players[ptype]:
                    self._synth.player_add(ptype, name, player)
        # ladspa effects
        wanted = self.bank[patch]["ladspafx"] | PATCHCORD
        if set(wanted) != set(self._synth.ladspafx):
            self._synth.fxchain_clear()
            for name, fx in wanted.items():
                self._synth.fxchain_add(name, fx)
            self._synth.fxchain_connect()
        # counters
        for name in list(self._router.counters):
            if name not in self.bank[patch]["counters"]:
                del self._router.counters[name]
        for name, counter in self.bank[patch]["counters"].items():
            if name not in self._router.counters:
                self._router.counters[name] = counter
                counter.val = counter.startval
            if name in self.bank.patch[patch].get("counters", {}):
                counter.val = counter.startval
        # midi rules
        self._router.reset()
        for rule in self.bank[patch]["rules"]:
            self.add_midirule(rule)
        # midi messages
        for msg in self.bank[patch]["messages"]:
            self.send_midimessage(msg)

    def update_patch(self, name):
        """
        Write current synth state back into a patch.

        Args:
          name (str): Patch name to modify in-place.
        """
        self.bank.patch[name]["messages"] = []
        sfonts = {self._sfonts[sf].id: sf for sf in self._sfonts}
        for chan in range(1, self._synth["synth.midi-channels"] + 1):
            for cc, default in enumerate(CC_DEFAULTS):
                val = self._synth.get_cc(chan, cc)
                if val != default and default != -1 :
                    self.bank.patch[name]["messages"].append(
                        MidiMessage(type="cc", chan=chan, num=cc, val=val)
                    )
            id, bank, prog = self._synth.program_info(chan)
            if id not in sfonts:
                if chan in self.bank.patch[name]:
                    del self.bank.patch[name][chan]
            else:
                self.bank.patch[name][chan] = SFPreset(sfonts[id], bank, prog)

    def add_midirule(self, rule):
        """
        Install a live MIDI rule after current bank rules

        Args:
          rule (MidiRule): A temporary rule
        """
        self._router.add(rule)

    def send_midimessage(self, msg):
        """
        Send a MIDI message directly to the synth.

        Args:
          msg (MidiMessage): MIDI message object.
        """
        self._synth.send_midievent(msg)

    def fluidsetting(self, name):
        """
        Retrieve a FluidSynth setting.

        Args:
          name (str): A FluidSynth setting name (e.g. 'synth.gain').

        Returns:
          (int|float|str): Current value of the setting.
        """
        return self._synth[name]
        
    def fluidsetting_set(self, name, val):
        """
        Modify a FluidSynth setting.

        Args:
          name (str): Setting name (only 'synth.' prefixes allowed).
          val (int|float|str): Desired value.
        """
        self._synth[name] = val

    def set_callback(self, func):
        """
        Install a callback to observe MIDI events.

        Args:
          func (callable | None):
              Function taking a single event, or None to disable.
        """
        if func:
            self._router.callback = func
        else:
            self._router.callback = lambda event: None

